#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "ising.h"


void ising_make_common(ising_model *G)
{
    int *numneighs; /* tmp var to help construct neigh */
    int i;
    int s1, s2;

    /* initialise with question-marks */
    for (i=0;i<(G->n);i++) {
        G->spins[i]=JOKER;
    }
    
    /* compute neigh list from edges (doesn't assume anything about how
     * G->edges was constructed, but should be /reasonably/ efficient if sparse) */
    numneighs=calloc(G->n,sizeof(int));
    for (i=0;i<(G->e);i++) {
        numneighs[G->edges[i].i]++;
        numneighs[G->edges[i].j]++;
        /* fprintf(stderr,"(%d -> %d)\n",G->edges[i].i,G->edges[i].j);
         * */
    }
    /*
    for (i=0;i<(G->n);i++)
        fprintf(stderr,"Node %d has %d neighbors\n",i,numneighs[i]);
    */
    G->neighidx[0]=0;
    for (i=1;i<(G->n);i++)
        G->neighidx[i]=G->neighidx[i-1]+numneighs[i-1]+1;
    assert(G->neighidx[G->n-1]==(2*G->e+G->n-numneighs[G->n-1]-1));
    for (i=0;i<(G->e);i++) {
        s1=G->edges[i].i;
        G->neigh[G->neighidx[s1]]++;
        G->neigh[G->neighidx[s1]+G->neigh[G->neighidx[s1]]]=i;
        
        s2=G->edges[i].j;
        G->neigh[G->neighidx[s2]]++;
        G->neigh[G->neighidx[s2]+G->neigh[G->neighidx[s2]]]=i;
    }
    free(numneighs);
}


/*
 * Allocates memory and puts down bonds to create Graph structure.
 *
 * Raster order is same as in Fortran
 */
ising_model* ising_make(int rows, int cols, int type)
{
    ising_model *G=malloc(sizeof(ising_model));
    int i,j;
    int l=0;
    
    G->beta=1;
    G->n=rows*cols;
    G->rows=rows; /*                                     */
    G->cols=cols; /* These three just for plotting later */
    G->type=type; /*                                     */
    switch (type) {
        case 0:
            G->e=G->n*2-rows-cols;
            break;
        case 1:
            G->e=G->n*2;
            break;
        case 2:
            G->e=(G->n)*(G->n-1)/2;
            break;
        case 3:
            G->e=(G->n)*3;
            break;
        default:
            /* Undefined type */
            printf("Unsupported topology, type %d\n", type);
            free(G);
            abort();
    }
    G->edges=calloc(G->e,sizeof(ising_edge));
    G->h=calloc(G->n,sizeof(double));
    G->spins=malloc(sizeof(int)*G->n);
    G->neigh=calloc(2*G->e+G->n,sizeof(int));
    G->neighidx=malloc(sizeof(int)*G->n);
    
    
    switch (type)
    {
        case 0:
            /* Rectangle, no wrap-around */
            for (i=0;i<(G->n);i++) {
                if ((i%rows)!=0) {
                    G->edges[l].i=i;
                    G->edges[l++].j=i-1;
                }
                if (i>=rows) {
                    G->edges[l].i=i;
                    G->edges[l++].j=i-rows;
                }
            }
            break;
        case 1:
            /* Rectangle, toroidal boundaries */
            for (i=0;i<(G->n);i++) {
                G->edges[l].i=i;
                if ((i%rows)==(rows-1))
                    G->edges[l++].j=i+1-rows;
                else
                    G->edges[l++].j=i+1;
                G->edges[l].i=i;
                G->edges[l++].j=(i+rows)%G->n;
            }
            break;
        case 2:
            /* Dense graph */
            for (i=0;i<(G->n);i++) {
                for (j=0;j<(G->n);j++) {
                    if (i<j) {
                        G->edges[l].i=i;
                        G->edges[l++].j=j;
                    }
                }
            }
            break;
        case 3:
            /* Triangular lattice, toroidal boundaries */
            for (i=0;i<(G->n);i++) {
                /* horizontal */
                G->edges[l].i=i;
                G->edges[l++].j=(i+rows)%G->n;
                
                /* always grab very next node down */
                G->edges[l].i=i;
                G->edges[l++].j=(i+1)%G->n;

                /* the other one */
                G->edges[l].i=i;
                if (i%rows%2)
                    G->edges[l++].j=(i-rows+1)%G->n;
                else
                    G->edges[l++].j=(i+rows+1)%G->n;
            }
            break;
        default:
            ising_free(G);
            abort();
    }
    assert(l==G->e);
    
    /* Just set all edges to 1 for now (biases left at 0) */
    for (i=0;i<(G->e);i++)
        G->edges[i].J=1;
    
    ising_make_common(G);
    
    return G;
}


ising_model* ising_clone(const ising_model* G1)
{
    int i;
    ising_model *G2=malloc(sizeof(ising_model));
    G2->beta=G1->beta;
    G2->n=G1->n;
    G2->e=G1->e;
    G2->rows=G1->rows;
    G2->cols=G1->cols;
    G2->type=G1->type;
    G2->edges=malloc(sizeof(ising_edge)*G1->e);
    G2->h=malloc(sizeof(double)*G1->n);
    G2->spins=malloc(sizeof(int)*G1->n);
    G2->neigh=malloc(sizeof(int)*(G1->n+2*G1->e));
    G2->neighidx=malloc(sizeof(int)*G1->n);
    for (i=0;i<G1->n;i++) {
        G2->h[i]=G1->h[i];
        G2->spins[i]=G1->spins[i];
        G2->neighidx[i]=G1->neighidx[i];
    }
    for (i=0;i<G1->e;i++)
        G2->edges[i]=G1->edges[i];
    for (i=0;i<(G1->n+2*G1->e);i++)
        G2->neigh[i]=G1->neigh[i];
    return G2;
}

void ising_free(ising_model *grid)
{
    free(grid->edges);
    free(grid->h);
    free(grid->spins);
    free(grid->neigh);
    free(grid->neighidx);
    free(grid);
}


void ising_conditional(ising_model* G, int node, double* up_prob)
{
    int i;
    int *spins=G->spins;
    int *neigh=G->neigh;
    ising_edge *edges=G->edges;
    ising_edge edge;
    int idx = G->neighidx[node];
    int num_neigh = neigh[idx];
    double Lpstar_up = 0;
    spin cur_setting = ising_get_spin(G, node);

    /* Temporarily set node to UP, record log-pstar and set back */
    ising_set_spin(G, node, UP);
    Lpstar_up += G->h[node];
    for(i=0; i<num_neigh; ++i) {
        idx++;
        edge=edges[neigh[idx]];
        Lpstar_up += ((spins[edge.i])*(spins[edge.j]))*edge.J;
    }
    ising_set_spin(G, node, cur_setting);

    /* Note Lpstar_down = -Lpstar_up, which gives: */
    (*up_prob) = 1.0/(1.0+exp(-2*(G->beta)*Lpstar_up));
} 



/* Below here are candidates for moving out into the header file and inlining */

spin ising_get_spin(const ising_model* ising, int node)
{
    return ising->spins[node];
}

void ising_set_spin(ising_model* ising, int node, spin value)
{
    ising->spins[node] = value;
}


/* DANGER: only colors copied: assumes grids are both created and memory all
 * allocated and same size and topology. Params not copied. */
void ising_copy_colors(const ising_model *from, ising_model *to)
{
    int i;
    /* TODO check grids are actually same size and stuff */

    for (i=0;i<(from->n);i++)
        to->spins[i]=from->spins[i];
}


void ising_read_in_spins(ising_model* ising, const char* filename)
{
    int value, i, ret;
    FILE* fid = fopen(filename,"r");
    for (i=0; i < ising->n; ++i) {
        ret = fscanf(fid,"%d", &value);
        if (ret == EOF)
            abort();
        ising_set_spin(ising, i, value);
    }
    fclose(fid);
}


void ising_write_out_spins(const ising_model* ising, const char* filename)
{
    int value, i;
    FILE* fid = fopen(filename, "w");
    for (i=0; i < ising->n; ++i) {
        value = ising_get_spin(ising, i);
        fprintf(fid,"%d\n", value);
    }
    fclose(fid);
}


double find_E(ising_model* G)
{
    int i;
    double E = 0;
    ising_edge edge;
    ising_edge* edges = G->edges;
    double* h = G->h;
    int* spins = G->spins;

    for (i=0; i < G->e; i++) {
        edge = edges[i];
        E += spins[edge.i] * spins[edge.j] * edge.J;
    }
    for (i=0; i < G->n; i++)
        E += spins[i] * h[i];
    
    return E;
}

