#include <stdlib.h>
#include <stdio.h>

#include "ising_mavm_clamp.h"


void usage()
{
    fprintf(stderr, "Usage: ./ising_pm mode mavmK num_steps step_size\n");
    fprintf(stderr, "   eg: ./ising_pm a 0 1000 0.01\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Possible modes:\n");
    fprintf(stderr, "    a - standard pseudo-marginal MH; jointly propose u and theta.\n");
    fprintf(stderr, "    b - independence proposals for u; separate MH proposals for theta.\n");
    fprintf(stderr, "    c - independence proposals for u; separate slice proposals for theta.\n");
    fprintf(stderr, "    d - MH proposals for u; separate MH proposals for theta.\n");
    fprintf(stderr, "    e - slice moves for u; separate MH proposals for theta.\n");
    fprintf(stderr, "    f - slice moves for u; separate slice moves for theta.\n");
    fprintf(stderr, "    g - MH proposals for u; separate slice moves for theta.\n");
}


int main(int argc, char **argv)
{
    int i, K;
    int rows=10, cols=30, wrap=1;
    ising_model* ising;
    ising_model* ising2; 
    ising_model* fit_ising;
    ising_model* dummy_ising;
    pmc_mavm_setup X;
    Rand_t seed;
    int S_iters;
    char mode;

    RanInit(seed, -1); /* From clock. TODO: allow changing */

    /* Command-line args */
    if (argc != 5) {
        usage();
        exit(1);
    }
    mode = argv[1][0];
    K = atof(argv[2]);
    S_iters = atof(argv[3])*2; /* *2 so get update for h and J per "iter" */
    GLOBAL_step_size = atof(argv[4]);

    /* Set up model and data */
    ising = ising_make(rows, cols, wrap);
    for (i=0; i < ising->e; ++i)
        ising->edges[i].J = 0.3; /* 0.3; */
    ising_read_in_spins(ising, "ISING_SPIN_VALUES");
    ising2 = ising_clone(ising);
    fit_ising = ising_clone(ising);
    dummy_ising = ising_clone(ising);

    /* Set up mavm_clamp structure for ising model */
    X.model = ising;
    X.prop_model = ising2;
    X.fit_model = fit_ising;
    X.dummy_model = dummy_ising;
    X.N_data = 1;
    X.prop_params = ising_prop_fn;
    X.slice_set_x = ising_slice_set_x;
    X.slice_get_x = ising_slice_get_x;
    X.Lprior = ising_prior_fn;
    X.Llike = (model_fn) find_E;
    X.report = ising_report_fn;
    X.param_mix = ising_param_mix;
    X.K_mix_steps = K;
    X.blanker = ising_ss_blanker;
    X.transition = ising_tr_function;
    X.num_u_in_transition = ising->n;
    X.has_coalesced = ising_ss_coalesced;

    switch (mode) {
        case 'a':
            pm_mavm(&X, S_iters, seed);
            break;
        case 'b':
            pmc_mavm(&X, S_iters, seed);
            break;
        case 'c':
            pmc_slice_mavm(&X, S_iters, seed);
            break;
        case 'd':
            pmc_mh_mh_mavm(&X, S_iters, seed);
            break;
        case 'e':
            pmc_slice_mh_mavm(&X, S_iters, seed);
            break;
        case 'f':
            pmc_slice_slice_mavm(&X, S_iters, seed);
            break;
        case 'g':
            pmc_mh_slice_mavm(&X, S_iters, seed);
            break;
        default:
            fprintf(stderr, "\n *** Unknown mode ***\n\n");
            usage();
            exit(1);
    }

    ising_free(ising);
    ising_free(ising2);
    ising_free(fit_ising);
    ising_free(dummy_ising);
    return 0;
}

