#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>

#include "mavm_clamp.h"
#include "maxmin.h"
#include "slice_sample.h"

typedef struct RS
{
    double* uu;         /* arbitrary length */
    int len_defined;    /* uu_cftp contains junk after this */
    int len_allocated;  /* uu_cftp needs reallocing to write in numbers beyond this */
    int len_last_used;  /* uu_cftp looked at on last run */
    int gaussian;       /* bool flag, whether to generate Gaussians rather than Uniforms */
    int combining;      /* bool flag, whether to slave off two other stores */
    double step_size;   /* How far along direction from rs2 to step. */
    struct RS* rs1;     /* if combining, assume rs1 is uniform and rs2 is Gaussian: */
    struct RS* rs2;     /* uu = (rs1->uu) + step_size*(rs2->uu) -- with reflection back into box*/
} random_store;

random_store* create_u_store()
{
    random_store* u_store;
    u_store = malloc(sizeof(random_store));
    u_store->uu = NULL;
    u_store->len_defined = 0;
    u_store->len_allocated = 0;
    u_store->len_last_used = 0;
    u_store->combining = 0;
    u_store->gaussian = 0;
    return u_store;
}

void free_u_store(random_store* u_store)
{
    free(u_store->uu);
    free(u_store);
}

void reset_u_store(random_store* u_store)
{
    u_store->len_defined = 0;
    u_store->len_last_used = 0;
}

double add_reflect(double xx, double direction, double scale)
{
    double target;
    double ipart;
    double fpart;
    int is_even;
    /* If this stuff becomes a bottle-neck I'm sure it could be done faster with
     * bit twiddling and/or a better choice of standard functions. */
    target = fabs(xx + scale*direction);
    ipart = floor(target);
    fpart = target - ipart;
    is_even = (ipart - floor(ipart/2)*2) < 0.1;
    if (is_even)
        return fpart;
    else
        return (1 - fpart);
}

void inform_u_store_of_need(random_store* u_store, int len, Rand_t seed)
    /* Tell the u_store that I'm about to access up to len numbers.
     * Keep a note of how much I've peeked at.
     * If necessary allocate more memory, and or generate more random numbers. */
{
    int i;

    u_store->len_last_used = max(u_store->len_last_used, len);
    if (len > (u_store->len_allocated)) {
        u_store->uu = (double*) realloc(u_store->uu, len*sizeof(double));
        if (u_store->uu == NULL) {
            fprintf(stderr, "Not enough memory to realloc. Bombing out.\n");
            abort();
        }
        u_store->len_allocated = len;
    }
    if (len > (u_store->len_defined)) {
        if (u_store->combining) {
            /* explicitly combine more of child streams */
            inform_u_store_of_need(u_store->rs1, len, seed);
            inform_u_store_of_need(u_store->rs2, len, seed);
            for (i = u_store->len_defined; i < len; i++) {
                u_store->uu[i] = add_reflect(
                        u_store->rs1->uu[i],
                        u_store->rs2->uu[i],
                        u_store->step_size);
            }
        } else {
            /* retrospectively generate previously unseen randomness */
            if (u_store->gaussian) {
                for (i = u_store->len_defined; i < len; i++) {
                    u_store->uu[i] = Rangauss(seed);
                }
            } else {
                for (i = u_store->len_defined; i < len; i++) {
                    u_store->uu[i] = Randouble(seed);
                }
            }
        }
        u_store->len_defined = len;
    }
}

int cftp_with_storage(
        void* summary_state,
        ss_blanker blanker,
        transition_function transition,
        size_t num_u_in_transition,
        ss_coalesced has_coalesced,
        random_store* u_store,
        Rand_t seed)
    /* Returns number of transitions performed */
{
    int i, sweep, total_sweeps;
    size_t num_u, num_sweeps;
    double* uu;

    total_sweeps = 0;
    u_store->len_last_used = 0;
    for (i=0; i < MAX_DOUBLINGS; i++) {
        /* fprintf(stderr, "Chunk i=%d of CFTP\n", i); */
        /* Initialize with question marks */
        blanker(summary_state);
        /* Make sure have enough random numbers (reusing ones already defined) */
        num_sweeps = INIT_ITERS * (1<<i);
        num_u = num_sweeps * num_u_in_transition;
        inform_u_store_of_need(u_store, num_u, seed);
        /* simulate from past to present (use uu backwards) */
        uu = u_store->uu + num_u;
        for (sweep=0; sweep < num_sweeps; sweep++) {
            uu -= num_u_in_transition;
            transition(summary_state, uu);
            total_sweeps++;
        }
        assert(uu == u_store->uu);
        /* Coalesced? */
        if (has_coalesced(summary_state))
            return total_sweeps;
    }
    fprintf(stderr,"You waited 'till the end of the universe, but it\n");
    fprintf(stderr,"still didn't finish. Sorry!\n");
    abort();
}


double pmc_mavm_log_score(
        pmc_mavm_setup* X,
        random_store* cftp_u_store,
        random_store* refine_u_store,
        Rand_t seed)
{
    int i, j, num_refine_u;
    double Lfhat;
    double* uu;
    int K;

    Lfhat = X->Lprior(X->model);
    if (Lfhat <= -DBL_MAX)
        return Lfhat;

    K = X->K_mix_steps;

    num_refine_u = X->N_data * K * X->num_u_in_transition;
    inform_u_store_of_need(refine_u_store, num_refine_u, seed);
    uu = refine_u_store->uu;
    for (i = 0; i < X->N_data; i++) {
        X->param_mix(X->dummy_model, X->model, X->fit_model, 1);
        /* (*mc_counter) = (*mc_counter) + */
        cftp_with_storage(
                X->dummy_model,
                X->blanker,
                X->transition,
                X->num_u_in_transition,
                X->has_coalesced,
                cftp_u_store,
                seed);
        for (j = K + 1; j > 1; j--) {
            Lfhat -= X->Llike(X->dummy_model);
            X->param_mix(X->dummy_model, X->model, X->fit_model, (j - 1.0) / (K + 1.0));
            Lfhat += X->Llike(X->dummy_model);
            X->transition(X->dummy_model, uu);
            uu += X->num_u_in_transition;
            /* (*mc_counter) = (*mc_counter) + 1; */
        }
        Lfhat -= X->Llike(X->dummy_model);
        X->param_mix(X->dummy_model, X->model, X->fit_model, 0);
        Lfhat += X->Llike(X->dummy_model);
    }
    assert(uu == (refine_u_store->uu + num_refine_u));

    Lfhat += X->Llike(X->model);

    return Lfhat;
}


void pm_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
{
    void* tmp_model;
    double MH_Lratio = 0;
    double prop_Lratio = 0;
    double Lfh, Lfh_prop;
    int ss;
    random_store* cftp_u_store;
    random_store* refine_u_store;

    cftp_u_store = create_u_store();
    refine_u_store = create_u_store();

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        prop_Lratio = X->prop_params(X->prop_model, X->model, seed);
        /* Swap in the proposed model */
        tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;

        /* Evaluate proposed model, using fresh random numbers */
        reset_u_store(cftp_u_store);
        reset_u_store(refine_u_store);
        Lfh_prop = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);

        MH_Lratio = Lfh_prop - Lfh + prop_Lratio;

        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, leave proposed model in place */
            Lfh = Lfh_prop;
        } else {
            /* reject, swap proposed model back out */
            tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;
        }
        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(refine_u_store);
}


void pmc_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
{
    void* tmp_model;
    double MH_Lratio = 0;
    double prop_Lratio = 0;
    double Lfh, Lfh_prop;
    int ss;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    random_store* cftp_u_store_prop;
    random_store* refine_u_store_prop;
    random_store* tmp_u_store;

    cftp_u_store = create_u_store();
    refine_u_store = create_u_store();
    cftp_u_store_prop = create_u_store();
    refine_u_store_prop = create_u_store();

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        /* Propose changing u_store */
        /* ------------------------ */
        reset_u_store(cftp_u_store_prop);
        reset_u_store(refine_u_store_prop);
        Lfh_prop = pmc_mavm_log_score(X, cftp_u_store_prop, refine_u_store_prop, seed);
        MH_Lratio = Lfh_prop - Lfh;
        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, swap in new u_stores */
            Lfh = Lfh_prop;
            tmp_u_store = cftp_u_store; cftp_u_store = cftp_u_store_prop; cftp_u_store_prop = tmp_u_store;
            tmp_u_store = refine_u_store; refine_u_store = refine_u_store_prop; refine_u_store_prop = tmp_u_store;
        }
        /* (don't need to do anything to reject here) */

        /* Proposed changing params */
        /* ------------------------ */
        prop_Lratio = X->prop_params(X->prop_model, X->model, seed);
        /* Swap in the proposed model */
        tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;

        /* Evaluate proposed model, using fresh random numbers */
        Lfh_prop = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);

        MH_Lratio = Lfh_prop - Lfh + prop_Lratio;

        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, leave proposed model in place */
            Lfh = Lfh_prop;
        } else {
            /* reject, swap proposed model back out */
            tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;
        }
        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(refine_u_store);
    free_u_store(cftp_u_store_prop);
    free_u_store(refine_u_store_prop);
}


typedef struct
{
    pmc_mavm_setup* X;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    unsigned* seed;
} mavm_slice_stuff;

double mavm_slice_fn(double xx, void* stuff)
{
    mavm_slice_stuff* S = stuff;
    pmc_mavm_setup* X = S->X;

    X->slice_set_x(xx, X->model);
    return pmc_mavm_log_score(
            X,
            S->cftp_u_store,
            S->refine_u_store,
            S->seed);
}

void pmc_slice_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
{
    double MH_Lratio = 0;
    double Lfh, Lfh_prop, xx;
    mavm_slice_stuff slice_stuff;
    int ss;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    random_store* cftp_u_store_prop;
    random_store* refine_u_store_prop;
    random_store* tmp_u_store;

    cftp_u_store = create_u_store();
    refine_u_store = create_u_store();
    cftp_u_store_prop = create_u_store();
    refine_u_store_prop = create_u_store();

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        /* Propose changing u_store */
        /* ------------------------ */
        reset_u_store(cftp_u_store_prop);
        reset_u_store(refine_u_store_prop);
        Lfh_prop = pmc_mavm_log_score(X, cftp_u_store_prop, refine_u_store_prop, seed);
        MH_Lratio = Lfh_prop - Lfh;
        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, swap in new u_stores */
            Lfh = Lfh_prop;
            tmp_u_store = cftp_u_store; cftp_u_store = cftp_u_store_prop; cftp_u_store_prop = tmp_u_store;
            tmp_u_store = refine_u_store; refine_u_store = refine_u_store_prop; refine_u_store_prop = tmp_u_store;
        }
        /* (don't need to do anything to reject here) */

        /* Slice update for params */
        /* ----------------------- */
        /* TODO could rewrite MH moves in terms of same get_x and set_x used by
         * slice and make proposal mechanism generic, so not rewriting for
         * future models */
        xx = X->slice_get_x(X->model);
        slice_stuff.X = X;
        slice_stuff.cftp_u_store = cftp_u_store;
        slice_stuff.refine_u_store = refine_u_store;
        slice_stuff.seed = &(seed[0]);
        slice_update(xx, &slice_stuff, mavm_slice_fn, GLOBAL_step_size, &Lfh, seed);

        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(refine_u_store);
    free_u_store(cftp_u_store_prop);
    free_u_store(refine_u_store_prop);
}


void u_store_combination_accept_rejig(random_store* U)
    /* Stuff to do after accepting change given by u_store combination to put
     * new stream in primary place and reset combination stream. */ 
{
    random_store* rs1;
    double* uu_tmp;
    int la_tmp;

    assert(U->combining);
    rs1 = U->rs1;

    uu_tmp = rs1->uu;
    la_tmp = rs1->len_allocated;

    /* Transplant combination stream into primary u_store */
    rs1->uu = U->uu;
    rs1->len_defined =   U->len_defined;
    rs1->len_allocated = U->len_allocated;
    rs1->len_last_used = U->len_last_used;

    /* steal old primary u_store's memory for next combination */
    U->uu = uu_tmp;
    U->len_allocated = la_tmp;
    U->len_defined = 0;
    U->len_last_used = 0;

    /* reset direction stream */
    reset_u_store(U->rs2);
}

void u_store_combination_reject_rejig(random_store* U)
    /* Stuff to do after rejection change given by u_store combination,
     * leave primary stream in place, but reset combination and direction. */ 
{
    reset_u_store(U);
    reset_u_store(U->rs2);
}

void pmc_mh_mh_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
    /* MH sample u (fixed made up step size of 0.1), and do mh for theta using
     * user-specified step-size. */
{
    void* tmp_model;
    double MH_Lratio = 0;
    double prop_Lratio = 0;
    double Lfh, Lfh_prop;
    int ss;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    random_store* direction_cftp;
    random_store* combine_cftp;
    random_store* direction_refine;
    random_store* combine_refine;

    cftp_u_store = create_u_store();
    direction_cftp = create_u_store();
    direction_cftp->gaussian = 1;
    combine_cftp = create_u_store();
    combine_cftp->combining = 1;
    combine_cftp->step_size = 0.1;
    combine_cftp->rs1 = cftp_u_store;
    combine_cftp->rs2 = direction_cftp;

    refine_u_store = create_u_store();
    direction_refine = create_u_store();
    direction_refine->gaussian = 1;
    combine_refine = create_u_store();
    combine_refine->combining = 1;
    combine_refine->step_size = 0.1;
    combine_refine->rs1 = refine_u_store;
    combine_refine->rs2 = direction_refine;

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        /* Propose changing u_store */
        /* ------------------------ */
        Lfh_prop = pmc_mavm_log_score(X, combine_cftp, combine_refine, seed);
        MH_Lratio = Lfh_prop - Lfh;
        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, swap in new u_stores */
            Lfh = Lfh_prop;
            u_store_combination_accept_rejig(combine_cftp);
            u_store_combination_accept_rejig(combine_refine);
        } else {
            u_store_combination_reject_rejig(combine_cftp);
            u_store_combination_reject_rejig(combine_refine);
        }

        /* Proposed changing params */
        /* ------------------------ */
        prop_Lratio = X->prop_params(X->prop_model, X->model, seed);
        /* Swap in the proposed model */
        tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;

        /* Evaluate proposed model, using fresh random numbers */
        Lfh_prop = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);

        MH_Lratio = Lfh_prop - Lfh + prop_Lratio;

        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, leave proposed model in place */
            Lfh = Lfh_prop;
        } else {
            /* reject, swap proposed model back out */
            tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;
        }
        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(direction_cftp);
    free_u_store(combine_cftp);
    free_u_store(refine_u_store);
    free_u_store(direction_refine);
    free_u_store(combine_refine);
}


double mavm_uslice_fn(double xx, void* stuff)
{
    random_store* cftp_u;
    random_store* refine_u;
    mavm_slice_stuff* S = stuff;
    pmc_mavm_setup* X = S->X;

    /* Reset u_stores, but NOT their children, so have same central u's and
     * direction. Just change step size and let them recalculate as needed. */
    cftp_u = S->cftp_u_store;
    refine_u = S->refine_u_store;
    assert(cftp_u->combining);
    assert(refine_u->combining);
    reset_u_store(cftp_u);
    reset_u_store(refine_u);
    cftp_u->step_size = xx;
    refine_u->step_size = xx;

    return pmc_mavm_log_score(
            X,
            cftp_u,
            refine_u,
            S->seed);
}


void pmc_slice_mh_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
    /* Slice sample u, and do mh for theta */
{
    void* tmp_model;
    double MH_Lratio = 0;
    double prop_Lratio = 0;
    double Lfh, Lfh_prop;
    int ss;
    mavm_slice_stuff slice_stuff;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    random_store* direction_cftp;
    random_store* combine_cftp;
    random_store* direction_refine;
    random_store* combine_refine;

    cftp_u_store = create_u_store();
    direction_cftp = create_u_store();
    direction_cftp->gaussian = 1;
    combine_cftp = create_u_store();
    combine_cftp->combining = 1;
    combine_cftp->step_size = 0.1;
    combine_cftp->rs1 = cftp_u_store;
    combine_cftp->rs2 = direction_cftp;

    refine_u_store = create_u_store();
    direction_refine = create_u_store();
    direction_refine->gaussian = 1;
    combine_refine = create_u_store();
    combine_refine->combining = 1;
    combine_refine->step_size = 0.1;
    combine_refine->rs1 = refine_u_store;
    combine_refine->rs2 = direction_refine;

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        /* Propose changing u_store */
        /* ------------------------ */
        slice_stuff.X = X;
        slice_stuff.cftp_u_store = combine_cftp;
        slice_stuff.refine_u_store = combine_refine;
        slice_stuff.seed = &(seed[0]);
        slice_update(0.0, &slice_stuff, mavm_uslice_fn, 1.0, &Lfh, seed);
        u_store_combination_accept_rejig(combine_cftp);
        u_store_combination_accept_rejig(combine_refine);

        /* Proposed changing params */
        /* ------------------------ */
        prop_Lratio = X->prop_params(X->prop_model, X->model, seed);
        /* Swap in the proposed model */
        tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;

        /* Evaluate proposed model, using fresh random numbers */
        Lfh_prop = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);

        MH_Lratio = Lfh_prop - Lfh + prop_Lratio;

        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, leave proposed model in place */
            Lfh = Lfh_prop;
        } else {
            /* reject, swap proposed model back out */
            tmp_model = X->model; X->model = X->prop_model; X->prop_model = tmp_model;
        }
        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(direction_cftp);
    free_u_store(combine_cftp);
    free_u_store(refine_u_store);
    free_u_store(direction_refine);
    free_u_store(combine_refine);
}


void pmc_slice_slice_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
    /* Slice sample u, and do mh for theta */
{
    double Lfh, xx;
    int ss;
    mavm_slice_stuff slice_stuff;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    random_store* direction_cftp;
    random_store* combine_cftp;
    random_store* direction_refine;
    random_store* combine_refine;

    cftp_u_store = create_u_store();
    direction_cftp = create_u_store();
    direction_cftp->gaussian = 1;
    combine_cftp = create_u_store();
    combine_cftp->combining = 1;
    combine_cftp->step_size = 0.1;
    combine_cftp->rs1 = cftp_u_store;
    combine_cftp->rs2 = direction_cftp;

    refine_u_store = create_u_store();
    direction_refine = create_u_store();
    direction_refine->gaussian = 1;
    combine_refine = create_u_store();
    combine_refine->combining = 1;
    combine_refine->step_size = 0.1;
    combine_refine->rs1 = refine_u_store;
    combine_refine->rs2 = direction_refine;

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        /* Propose changing u_store */
        /* ------------------------ */
        slice_stuff.X = X;
        slice_stuff.cftp_u_store = combine_cftp;
        slice_stuff.refine_u_store = combine_refine;
        slice_stuff.seed = &(seed[0]);
        slice_update(0.0, &slice_stuff, mavm_uslice_fn, 1.0, &Lfh, seed);
        u_store_combination_accept_rejig(combine_cftp);
        u_store_combination_accept_rejig(combine_refine);

        /* Slice update for params */
        /* ----------------------- */
        xx = X->slice_get_x(X->model);
        slice_stuff.X = X;
        slice_stuff.cftp_u_store = cftp_u_store;
        slice_stuff.refine_u_store = refine_u_store;
        slice_stuff.seed = &(seed[0]);
        slice_update(xx, &slice_stuff, mavm_slice_fn, GLOBAL_step_size, &Lfh, seed);

        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(direction_cftp);
    free_u_store(combine_cftp);
    free_u_store(refine_u_store);
    free_u_store(direction_refine);
    free_u_store(combine_refine);
}


void pmc_mh_slice_mavm(pmc_mavm_setup* X, int S_iters, Rand_t seed)
{
    double MH_Lratio = 0;
    double Lfh, Lfh_prop, xx;
    mavm_slice_stuff slice_stuff;
    int ss;
    random_store* cftp_u_store;
    random_store* refine_u_store;
    random_store* direction_cftp;
    random_store* combine_cftp;
    random_store* direction_refine;
    random_store* combine_refine;

    cftp_u_store = create_u_store();
    direction_cftp = create_u_store();
    direction_cftp->gaussian = 1;
    combine_cftp = create_u_store();
    combine_cftp->combining = 1;
    combine_cftp->step_size = 0.1;
    combine_cftp->rs1 = cftp_u_store;
    combine_cftp->rs2 = direction_cftp;

    refine_u_store = create_u_store();
    direction_refine = create_u_store();
    direction_refine->gaussian = 1;
    combine_refine = create_u_store();
    combine_refine->combining = 1;
    combine_refine->step_size = 0.1;
    combine_refine->rs1 = refine_u_store;
    combine_refine->rs2 = direction_refine;

    Lfh = pmc_mavm_log_score(X, cftp_u_store, refine_u_store, seed);
    for (ss=0; ss < S_iters; ss++) {
        /* Propose changing u_store */
        /* ------------------------ */
        Lfh_prop = pmc_mavm_log_score(X, combine_cftp, combine_refine, seed);
        MH_Lratio = Lfh_prop - Lfh;
        if (log(Randouble(seed)) < MH_Lratio) {
            /* accept, swap in new u_stores */
            Lfh = Lfh_prop;
            u_store_combination_accept_rejig(combine_cftp);
            u_store_combination_accept_rejig(combine_refine);
        } else {
            u_store_combination_reject_rejig(combine_cftp);
            u_store_combination_reject_rejig(combine_refine);
        }

        /* Slice update for params */
        /* ----------------------- */
        /* TODO could rewrite MH moves in terms of same get_x and set_x used by
         * slice and make proposal mechanism generic, so not rewriting for
         * future models */
        xx = X->slice_get_x(X->model);
        slice_stuff.X = X;
        slice_stuff.cftp_u_store = cftp_u_store;
        slice_stuff.refine_u_store = refine_u_store;
        slice_stuff.seed = &(seed[0]);
        slice_update(xx, &slice_stuff, mavm_slice_fn, GLOBAL_step_size, &Lfh, seed);

        X->report(X->model);
    }

    free_u_store(cftp_u_store);
    free_u_store(direction_cftp);
    free_u_store(combine_cftp);
    free_u_store(refine_u_store);
    free_u_store(direction_refine);
    free_u_store(combine_refine);
}


