function acf = autocorr_diag_rcoda(samples, lags)
%AUTOCORR_DIAG_RCODA estimates autocorrelations for each column (calls R-CODA)
%
%    acf = autocorr_diag_rcoda(samples, lags);
%
% This function simply shells R and runs the autocorr.diag routine in R-CODA.
%
% Inputs:
%     samples SxN each column of S samples is processed in isolation 
%        lags 1xK lags to compute autocorrelation at
%                 If ommitted get R-CODA's default: [0, 1, 5, 10, 50]
%
% Outputs:
%         acf KxN estimated autocorrelations for each column

% It's probably a good idea to have a native Matlab version of this routine.
% But then I would need this routine anyway to test it.

[S, N] = size(samples);
if (S == 1)
    error('Cannot process "time series" of length 1.');
end

% The user should put the version of R that they want to use in their system's
% PATH, but if that fails I will try looking in other places too.
Rlocs = {'', '/opt/local/bin/'};
for loc = Rlocs
    [status, output] = system([loc{:}, 'R --version']);
    if status == 0
        R_cmd = [loc{:}, 'R --vanilla CMD BATCH '];
        break;
    end
end
if ~exist('R_cmd')
    error('R executable not found');
end
% TODO cross-platform support for the following is untested. On my linux machine
% I need a clean environment or R fails. I do this with 'env -i'. I don't
% immediately know how to do this on other platforms, but maybe it isn't needed.
if exist('/usr/bin/env', 'file')
    R_cmd = ['/usr/bin/env -i ', R_cmd];
end

% Set up location for temporary files (in RAM if possible)
dirnm = ['acf', sprintf('%d', floor(rand*100000))];
if exist('/dev/shm', 'dir')
    % Linux
    tmp_dir = ['/dev/shm/', dirnm, '/'];
else
    % Should work on all platforms
    tmp_dir = [tempdir(), dirnm, filesep()];
end
success = mkdir(tmp_dir);
assert(success);

% Write samples out to temporary file
save([tmp_dir, 'samples'], 'samples', '-ascii');

% Run R-CODA's routine on that file
if nargin < 2 || isempty(lags)
    lags_arg = [];
else
    lags_arg = sprintf(', lags=c(%d', lags(1));
    if numel(lags) > 1
        lags_arg = [lags_arg, sprintf(',%d', lags(2:end))];
    end
    lags_arg = [lags_arg, ')'];
end
R_program = ['library(coda)\n',...
    'mcmcread=read.table("samples")\n',...
    'mcmcrun=cbind(mcmcread)\n',...
    'mcmcobj=mcmc(mcmcrun)\n',...
    'acf <- autocorr.diag(mcmcobj', lags_arg, ')\n',...
    'write.table(acf, file="acf", row.names=FALSE, col.names=FALSE)\n'];
    %'print(acf)\n',...
R_file = [tmp_dir, 'prog.r'];
fid = fopen(R_file, 'w');
fprintf(fid, R_program);
fclose(fid);
out_file = [tmp_dir, 'tmpRout'];
acf_file = [tmp_dir, 'acf'];
opwd = pwd();
cd(tmp_dir);
system_cmd = [R_cmd, R_file, ' ', out_file];
status = system(system_cmd);
cd(opwd);
assert(~status);

% grab result from output
acf = load(acf_file);

% Delete temporary directory
rmdir(tmp_dir, 's');

