function [xx, log_Px] = slice_sweep(logdist, xx, widths, step_out, log_Px, bailout, varargin)
%SLICE_SWEEP simple slice sampling update applied to each element of an array
%
%     [xx, log_Px] = slice_sweep(logdist, xx, widths, step_out, log_Px, bailout, varargin)
%
% Inputs:
%       logdist  @fn  function handle: log_Px = logdist(xx, varargin{:})
%            xx  Dx1  initial state (or any shape array with D elements)
%        widths  Dx1  or 1x1, step sizes for slice sampling
%          ...or Dx2  fixed initial bracket placement for each element. Must contain xx!
%      step_out  bool  set to true if widths may sometimes be far too small
%        log_Px  1x1  initial log probability up to a constant if already computed
%       bailout  1x1  If true, don't throw error if current point not acceptable.
%                     Use initial point and initial log_Px (not recomputed one)
%                     If in doubt, set to false, and understand what's going on
%                     if you get the error (a bug unless logdist is
%                     intentionally noisy).
%      varargin   -   any extra arguments are passed on to logdist
%
% Outputs:
%            xx  Dx1  position after update (same shape as input)
%        log_Px  1x1  final log probability up to a constant

% startup stuff
D = numel(xx);
if numel(widths) == 1
    widths = repmat(widths, D, 1);
end
if nargin < 5
    log_Px = logdist(xx, varargin{:});
end

fixed_bounds = isequal(size(widths), [D, 2]);

x_l = xx;
x_r = xx;
xprime = xx;

% Sweep through axes (simplest thing)
for dd = 1:D
    log_uprime = log(rand()) + log_Px;

    % Create a horizontal interval (x_l, x_r) enclosing xx
    if fixed_bounds
        x_l(dd) = widths(dd, 1);
        x_r(dd) = widths(dd, 2);
        assert((xx(dd) > x_l(dd)) && (xx(dd) < x_r(dd)));
        step_size = x_r(dd) - x_l(dd);
    else
        rr = rand();
        step_size = widths(dd);
        x_l(dd) = xx(dd) - rr*step_size;
        x_r(dd) = xx(dd) + (1-rr)*step_size;
    end
    if step_out
        % Typo in early editions of book. Book said compare to u, but it should say u'
        while logdist(x_l, varargin{:}) > log_uprime
            x_l(dd) = x_l(dd) - step_size;
        end
        while logdist(x_r, varargin{:}) > log_uprime
            x_r(dd) = x_r(dd) + step_size;
        end
    end

    % Inner loop:
    % Propose xprimes and shrink interval until good one found
    while 1
        xprime(dd) = rand()*(x_r(dd) - x_l(dd)) + x_l(dd);
        log_Px = logdist(xprime, varargin{:});
        if log_Px > log_uprime
            break % this is the only way to leave the while loop
        else
            % Shrink in
            if xprime(dd) > xx(dd)
                x_r(dd) = xprime(dd);
            elseif xprime(dd) < xx(dd)
                x_l(dd) = xprime(dd);
            else
                if bailout
                    log_Px = log_Px_orig;
                    break;
                else
                    error('BUG DETECTED: Shrunk to current position and still not acceptable.');
                end
            end
        end
    end
    xx(dd) = xprime(dd);
    x_l(dd) = xprime(dd);
    x_r(dd) = xprime(dd);
end

